(** These are barebones bindings for the LZ4 C api. The functions are declared external in
    the mli so that calls go directly to the C api, rather than through an OCaml stub.

    Note that none of the sizes supplied by the caller are verified and arguments are
    handed off to the C api as is. These bindings can serve as the basis for a more
    highlevel API for LZ4 -OR- when the caller does not want to pay the cost of additional
    checks. *)

open Core.Std


(** Returns the maximum size of the compression output for a specified input size *)
external compressBound: int -> int
  = "_LZ4_compressBound"


(** Ctx is a data structure internal to the LZ4 api and is used by the compressCtx and
    compress64kCtx functions. For a slight a speedup, one can allocate Ctx once and reuse
    it for multiple calls.

    Note: Typically the use of the Ctx functions do not buy us any peformance. Thsi is
    because the Ctx structure is typically stack allocated and not heap allocated. They
    become relevant only if macros in the LZ4 C source are changed to force HEAP_MODE
    allocation of the Ctx *)
type ctx

external create_ctx: unit -> ctx = "_create_ctx"

module Lz4_string : sig

  (** Compress source into dest. The argument isize is the part of source to compress,
      typically String.length. The dest buffer is expected to be large enough to hold the
      compression output. The function returns the number of bytes of compressed output
      and 0 on error. *)
  external compress: source:string -> dest:string
    -> isize:int -> int
      = "_LZ4_compress_string"

  (** High compression verison of the compress that trades CPU for greater compression *)
  external compressHC: source:string -> dest:string
    -> isize:int -> int
      = "_LZ4_compressHC_string"

  (** Compress source to dest. The user must specify the size of the dest available for
      compression output. If compression is acheivable the number of bytes produced are
      returned. On error, 0 is retuned.  *)
  external compress_limitedOutput: source:string -> dest:string
    -> isize:int -> maxOutputSize:int -> int
      = "_LZ4_compress_limitedOutput_string"

  (** Use this function for source strings that are <= 64k. (See notes on the Ctx data
      type) *)
  external compress64kCtx: ctx -> source:string -> dest:string
    -> isize:int -> maxOutputSize:int -> int
      = "_LZ4_compress64kCtx_string"

  (** Use this function for source strings of large sizes. (See notes on the Ctx data
      type) *)
  external compressCtx: ctx -> source:string -> dest:string
    -> isize:int -> maxOutputSize:int -> int
      = "_LZ4_compressCtx_string"

  (** Uncompress source into dest. No more than osize bytes will be written out. The
      function returns the number of bytes read from source. On success, the function
      returns the number of bytes read from source. On error, the function returns a
      negative number indicating the offset faulty source offset. *)
  external uncompress: source:string -> dest:string
    -> osize:int -> int
      = "_LZ4_uncompress_string"

  (** Uncompress source into dest. Input isize specifies the number of compressed bytes
      (typically (String.length source)). No more than maxOutputSize bytes will be written
      out. On success, the function returns the number of bytes produced in the ouptut. On
      error the function produces a negative number indicating the offset faulty source
      offset. *)
  external uncompress_unknownOutputSize: source:string -> dest:string
    -> isize:int -> maxOutputSize:int -> int
      = "_LZ4_uncompress_unknownOutputSize_string"

  (** Allocate a buffer large enough to hold the compressed output of the input. *)
  val make_compression_dest: string -> string
end


module Lz4_bigstring : sig

  external compress: source:Bigstring.t -> dest:Bigstring.t
    -> isize:int -> int
      = "_LZ4_compress_bigstring"

  external compressHC: source:Bigstring.t -> dest:Bigstring.t
    -> isize:int -> int
      = "_LZ4_compressHC_bigstring"

  external compress_limitedOutput: source:Bigstring.t -> dest:Bigstring.t
    -> isize:int -> maxOutputSize:int -> int
      = "_LZ4_compress_limitedOutput_bigstring"

  external compressCtx: ctx -> source:Bigstring.t -> dest:Bigstring.t
    -> isize:int -> maxOutputSize:int -> int
      = "_LZ4_compressCtx_bigstring"

  external compress64kCtx: ctx -> source:Bigstring.t -> dest:Bigstring.t
    -> isize:int -> maxOutputSize:int -> int
      = "_LZ4_compress64kCtx_bigstring"

  external uncompress: source:Bigstring.t -> dest:Bigstring.t
    -> osize:int -> int
      = "_LZ4_uncompress_bigstring"

  external uncompress_unknownOutputSize: source:Bigstring.t -> dest:Bigstring.t
    -> isize:int -> maxOutputSize:int -> int
      = "_LZ4_uncompress_unknownOutputSize_bigstring"

  val make_compression_dest : Bigstring.t -> Bigstring.t
end





